from src.chercher.cli import _prune
from .plugin_mocks import BadPlugin, DummyTxtPlugin


def test_prune_removes_correct_document(faker, test_db, plugin_manager):
    txt_uri = faker.file_path(depth=3, extension="txt")
    pdf_uri = faker.file_path(depth=3, extension="pdf")

    cursor = test_db.cursor()
    cursor.execute(
        "INSERT INTO documents (uri, title, body, hash, metadata) VALUES (?, ?, ?, ?, ?)",
        (txt_uri, "", "", "", "{}"),
    )
    cursor.execute(
        "INSERT INTO documents (uri, title, body, hash, metadata) VALUES (?, ?, ?, ?, ?)",
        (pdf_uri, "", "", "", "{}"),
    )
    test_db.commit()

    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()
    assert len(documents) == 2

    plugin_manager.register(DummyTxtPlugin())
    _prune(test_db, plugin_manager)

    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()
    assert len(documents) == 1


def test_prune_removes_correctly_all_document(faker, test_db, plugin_manager):
    cursor = test_db.cursor()
    for i in range(3):
        uri = faker.file_path(depth=i + 1, extension="txt")
        cursor.execute(
            "INSERT INTO documents (uri, title, body, hash, metadata) VALUES (?, ?, ?, ?, ?)",
            (uri, "", "", "", "{}"),
        )
        test_db.commit()

    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()
    assert len(documents) == 3

    plugin_manager.register(DummyTxtPlugin())
    _prune(test_db, plugin_manager)

    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()
    assert len(documents) == 0


def test_purge_with_exception(faker, test_db, plugin_manager):
    cursor = test_db.cursor()

    txt_uri = faker.file_path(depth=3, extension="txt")
    pdf_uri = faker.file_path(depth=3, extension="pdf")

    cursor = test_db.cursor()
    cursor.execute(
        "INSERT INTO documents (uri, title, body, hash, metadata) VALUES (?, ?, ?, ?, ?)",
        (txt_uri, "", "", "", "{}"),
    )
    cursor.execute(
        "INSERT INTO documents (uri, title, body, hash, metadata) VALUES (?, ?, ?, ?, ?)",
        (pdf_uri, "", "", "", "{}"),
    )
    test_db.commit()

    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()
    assert len(documents) == 2

    plugin_manager.register(BadPlugin())
    plugin_manager.register(DummyTxtPlugin())
    _prune(test_db, plugin_manager)

    cursor.execute("SELECT * FROM documents")
    documents = cursor.fetchall()
    assert len(documents) == 1



import sys
import sqlite3
import click
from loguru import logger
import pluggy
from chercher.utils import console
from chercher.output import print_plugins_table, print_results_list, print_results_table
from chercher.plugin_manager import get_plugin_manager
from chercher.settings import Settings, APP_NAME, APP_DIR
from chercher.db import init_db, db_connection


logger.remove()
logger.add(
    sys.stderr,
    colorize=True,
    format="<green>{time:HH:mm:ss}</green> | <level>{level}</level> | {message}",
    level="INFO",
)
logger.add(
    APP_DIR / "chercher_errors.log",
    rotation="10 MB",
    retention="15 days",
    level="ERROR",
)

settings = Settings()


@click.group(help=settings.description)
@click.version_option(
    version=settings.version,
    message="v%(version)s",
    package_name=APP_NAME,
)
@click.pass_context
def cli(ctx: click.Context) -> None:
    with db_connection(settings.db_url) as conn:
        logger.debug("initializing the database")
        init_db(conn)

    ctx.ensure_object(dict)
    ctx.obj["settings"] = settings
    ctx.obj["db_url"] = settings.db_url
    ctx.obj["pm"] = get_plugin_manager()


def _index(conn: sqlite3.Connection, uris: list[str], pm: pluggy.PluginManager) -> None:
    cursor = conn.cursor()
    plugin_settings = dict(settings).get("plugin", {})

    for uri in uris:
        try:
            for documents in pm.hook.ingest(uri=uri, settings=plugin_settings):
                for doc in documents:
                    try:
                        cursor.execute(
                            """
                    INSERT INTO documents (uri, title, body, hash, metadata) VALUES (?, ?, ?, ?, ?)
                    """,
                            (
                                doc.uri,
                                doc.title,
                                doc.body,
                                doc.hash,
                                doc.metadata.model_dump_json(),
                            ),
                        )
                        conn.commit()
                        logger.info(f'document "{doc.uri}" indexed')
                    except sqlite3.IntegrityError:
                        logger.warning(f'document "{doc.uri}" already exists')
                    except Exception as e:
                        logger.error(
                            f"something went wrong while indexing '{doc.uri}': {e}"
                        )
        except Exception as e:
            logger.error(
                f"something went wrong while trying to index documents from '{uri}': {e}"
            )


@cli.command(help="Index documents, webpages and more.")
@click.argument("uris", nargs=-1)
@click.pass_context
def index(ctx: click.Context, uris: list[str]) -> None:
    pm = ctx.obj["pm"]
    db_url = ctx.obj["db_url"]

    if not pm.list_name_plugin():
        logger.warning("No plugins registered!")
        return

    with db_connection(db_url) as conn:
        _index(conn, uris, pm)


def _prune(conn: sqlite3.Connection, pm: pluggy.PluginManager) -> None:
    cursor = conn.cursor()
    plugin_settings = dict(settings).get("plugin", {})

    try:
        cursor.execute("SELECT uri, hash FROM documents")
        uris_and_hashes = cursor.fetchall()
    except Exception as e:
        logger.error(
            f"something went wrong while retrieving documents from the database: {e}"
        )
        return

    for uri, hash in uris_and_hashes:
        try:
            for result in pm.hook.prune(uri=uri, hash=hash, settings=plugin_settings):
                if not result:
                    continue

                try:
                    cursor.execute("DELETE FROM documents WHERE uri = ?", (uri,))
                    conn.commit()
                    logger.info(f"document '{uri}' pruned")
                except Exception as e:
                    logger.error(f"something went wrong while purging '{uri}': {e}")
        except Exception as e:
            logger.error(
                f"something went wrong while trying to purge document '{uri}': {e}"
            )

    try:
        cursor.execute("VACUUM;")
        logger.info("vacuum completed successfully")
    except Exception as e:
        logger.error(f"something went wrong while performing vacuum operation: {e}")
        return

    logger.info("database cleanup completed")


@cli.command(help="Prune unnecessary documents from the database.")
@click.pass_context
def prune(ctx: click.Context) -> None:
    pm = ctx.obj["pm"]
    db_url = ctx.obj["db_url"]

    with db_connection(db_url) as conn:
        _prune(conn, pm)


@cli.command(help="Seach for documents matching your query.")
@click.argument("query")
@click.option(
    "-l",
    "--limit",
    type=int,
    default=5,
    help="Number of results.",
)
@click.option(
    "-o",
    "--output",
    type=click.Choice(["table", "list"], case_sensitive=False),
    default="table",
    help="Output format (available options: table and list).",
)
@click.pass_context
def search(ctx: click.Context, query: str, limit: int, output: str = "table") -> None:
    db_url = ctx.obj["db_url"]

    with db_connection(db_url) as conn:
        cursor = conn.cursor()

        sql_query = """
            SELECT uri, title, substr(body, 0, 300)
            FROM documents
            WHERE ROWID IN (
                SELECT ROWID
                FROM documents_fts
                WHERE documents_fts MATCH ?
                ORDER BY bm25(documents_fts)
                LIMIT ?
            )
            """

        cursor.execute(sql_query, (query, limit))
        results = cursor.fetchall()

        if not results:
            console.print(f"No results found for: '{query}'")
            return

        if output == "list":
            print_results_list(results)
        else:
            print_results_table(results)


@cli.command(help="List out all the registered plugins and their hooks.")
@click.pass_context
def plugins(ctx: click.Context) -> None:
    pm = ctx.obj["pm"]
    print_plugins_table(pm)



from pydantic import BaseModel, RootModel


class DocumentMetadata(RootModel):
    root: dict = {}


class Document(BaseModel):
    uri: str
    title: str | None = None
    body: str
    hash: str | None = None
    metadata: DocumentMetadata



from typing import Generator
from src.chercher import hookimpl, Document


class BadPlugin:
    @hookimpl
    def ingest(self, uri: str) -> Generator[Document, None, None]:
        if not uri.endswith(".txt"):
            raise Exception("Ooops!")

    @hookimpl
    def prune(self, uri: str) -> bool | None:
        if not uri.endswith(".txt"):
            raise Exception("Ooops!")


class DummyPlugin:
    @hookimpl
    def ingest(self, uri: str) -> Generator[Document, None, None]:
        yield Document(uri=uri, title=None, body="", hash=None, metadata={})

    @hookimpl
    def prune(self, uri: str) -> bool | None:
        return True


class DummyTxtPlugin:
    def __init__(
        self,
        title: str = "",
        body: str = "",
        hash: str = "",
        metadata: dict = {},
    ) -> None:
        self.title = title
        self.body = body
        self.hash = hash
        self.metadata = metadata

    @hookimpl
    def ingest(self, uri: str) -> Generator[Document, None, None]:
        if uri.endswith(".txt"):
            yield Document(
                uri=uri,
                body=self.body,
                title=self.title,
                hash=self.hash,
                metadata=self.metadata,
            )

    @hookimpl
    def prune(self, uri: str) -> bool | None:
        return uri.endswith(".txt")



def test_init_db_creates_required_tables(test_db):
    cursor = test_db.cursor()
    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='documents';"
    )
    documents_table = cursor.fetchone()
    assert documents_table is not None
    assert documents_table[0] == "documents"

    cursor.execute(
        "SELECT name FROM sqlite_master WHERE type='table' AND name='documents_fts';"
    )
    documents_fts_table = cursor.fetchone()
    assert documents_fts_table is not None
    assert documents_fts_table[0] == "documents_fts"


def test_document_insertion_saves_correctly(faker, test_db):
    uri = faker.file_path(depth=3)
    title = " ".join(faker.words())
    body = "\n".join(faker.sentences())
    hash = faker.sha256()

    cursor = test_db.cursor()
    cursor.execute(
        "INSERT INTO documents (uri, title, body, hash, metadata) VALUES (?, ?, ?, ?, ?)",
        (uri, title, body, hash, "{}"),
    )
    test_db.commit()

    cursor.execute("SELECT * FROM documents WHERE uri = ?", (uri,))
    document = cursor.fetchone()
    assert document[0] == uri
    assert document[1] == title
    assert document[2] == body
    assert document[3] == hash

    cursor.execute("SELECT * FROM documents_fts WHERE uri = ?", (uri,))
    document = cursor.fetchone()
    assert document[0] == uri
    assert document[1] == title
